import os

import numpy as np
import pandas as pd
from scipy import stats


def log_lhood_hist(theta, catalog, mc, dV, dt, t_shut, N_shut, PostRate):

    afb = theta[0]
    b = theta[1]
    tau = theta[2]

    Sigma = 10 ** (afb - b * mc)
    pred = np.cumsum(dV * np.diff(np.append(dt[0], dt))) * Sigma
    MetaSeismic = np.where(dt > t_shut)
    meta_t = dt - t_shut
    pred[MetaSeismic] = (-np.exp(
        -meta_t[MetaSeismic] / tau) + 1.) * PostRate * tau * Sigma + N_shut

    obs = np.interp(dt, catalog.ts.values, np.arange(0, len(catalog.ts)) + 1)
    obs_fac = [np.math.factorial(i) for i in np.diff(obs.astype(int))]
    return np.nansum(np.diff(obs) * np.log(np.diff(pred))
                     - np.diff(pred) - np.log(obs_fac))


def log_lhood_comp(theta, N, Mc, dV, dt, mag, Ts,
                   t_shut, N_shut, Q_shut, PostRate):

    af = theta[0]
    b = theta[1]
    tau = theta[2]

    A1 = N_shut * (af - b * Mc) / np.log10(np.exp(1))
    dV_ts = np.interp(Ts[Ts <= t_shut], dt[dt <= t_shut], dV[dt <= t_shut])
    dV_ts[dV_ts <= 0] = 1e-10
    K2 = np.nansum(np.log(dV_ts))
    K3 = (N - N_shut) * np.log(PostRate)
    A4 = -1 / tau * np.sum(Ts[Ts > t_shut] - t_shut)
    A5 = - 10 ** (af - b * Mc) * (Q_shut + PostRate * tau
                                  * (1 - np.exp(-(dt[-1] - t_shut) / tau)))
    A6 = N * np.log(b)
    K4 = N * np.log(np.log(10))
    A7 = -b * np.log(10) * np.sum(mag)
    A8 = N * b * (Mc - 0.005) * np.log(10)
    return -(A1 + K2 + K3 + A4 + A5 + A6 + K4 + A7 + A8)


def prior(afl=-12, afu=1, bl=0.5, bu=4.5, taul=1, tauu=2e5,
          daf=0.025, db=0.025, dtau=100, norm=False):
    # read values from SeismogenicIndces_prior.xlsx file
    df = pd.read_csv(os.path.dirname(__file__) + '/Priors.csv')
    af = df["afb1"].values
    af[np.isnan(af)] = df["afb2"].values[np.isnan(af)]

    b = df["b"].values

    tau = df["t (days)"].values * 86400
    tau[np.isnan(tau)] = np.nanmedian(tau)

    # b_prior
    b_a, b_b = stats.beta.fit((b - bl) / (bu - bl), floc=0, fscale=1)[0:2]
    b0 = np.arange(bl, bu, db)
    b_prior = stats.beta.pdf((b0 - bl) / (bu - bl), b_a, b_b)

    # af_prior
    af_a, af_b = stats.beta.fit(
        (af - afl) / (afu - afl), floc=0, fscale=1)[0:2]
    af0 = np.arange(afl, afu, daf)
    af_prior = stats.beta.pdf((af0 - afl) / (afu - afl), af_a, af_b)

    # tau_prior
    tau_a, bb, tau_b = stats.gamma.fit(tau, floc=0)
    tau0 = np.arange(taul, tauu, dtau)
    tau_prior = stats.gamma.pdf(tau0, tau_a, bb, tau_b)

    if norm:
        return af_prior / (np.sum(af_prior) * daf), \
            b_prior / (np.sum(b_prior) * db), \
            tau_prior / (np.sum(tau_prior) * dtau), \
            af0, b0, tau0, \
            daf, db, dtau
    else:
        return af_prior, b_prior, tau_prior, af0, b0, tau0


def log_lhood_comp_FF(theta, args):

    N, Mc, dV, dt, mag, Ts, t_shut, N_shut, Q_shut, PostRate = args
    A0 = theta[0]
    T0 = theta[1]
    b0 = theta[2]
    II = theta[3]

    LL = np.zeros((T0.shape[0], A0.shape[1], len(b0)))

    for j in range(0, len(b0)):
        A1 = N_shut * (A0 - b0[j] * Mc) / np.log10(np.exp(1))
        dV_ts = np.interp(Ts[Ts <= t_shut], dt[dt <= t_shut], dV[dt <= t_shut])
        dV_ts[dV_ts <= 0] = 1e-10
        K2 = np.nansum(np.log(dV_ts)) * II
        K3 = (N - N_shut) * np.log(PostRate) * II
        A4 = -1 / T0 * np.sum(Ts[Ts > t_shut] - t_shut)
        A5 = - 10 ** (A0 - b0[j] * Mc) * \
            (Q_shut + PostRate * T0 * (1 - np.exp(-(dt[-1] - t_shut) / T0)))
        A6 = N * np.log(b0[j]) * II
        K4 = N * np.log(np.log(10)) * II
        A7 = -b0[j] * np.log(10) * np.sum(mag) * II
        A8 = N * b0[j] * (Mc - 0.005) * np.log(10) * II
        LL[:, :, j] = (A1 + K2 + K3 + A4 + A5 + A6 + K4 + A7 + A8)

    return LL


def posterior(prior, LL):
    prior[0][prior[0] == 0] = 1e-10
    prior[1][prior[1] == 0] = 1e-10

    [X1, X2, X3] = np.meshgrid(np.log(prior[2]),
                               np.log(prior[0]),
                               np.log(prior[1]))
    log_f_j_p = X1 + X2 + X3

    temp = np.exp((LL - np.max(LL)) + log_f_j_p, LL)

    posterior = temp / np.sum(temp) * (1 / prior[6] / prior[7] / prior[8])

    return posterior
