import logging
from datetime import timedelta

import numpy as np
import scipy.optimize as opt
from hermes_model import ModelInput
from hermes_model.logging import set_up_logging
from hydws.parser import BoreholeHydraulics
from scipy import stats
from scipy.special import lambertw
from seismostats import Catalog, ForecastGRRateGrid
from seismostats.analysis.bvalue import estimate_b
from seismostats.analysis.estimate_mc import estimate_mc_maxc
from seismostats.utils.binning import bin_to_precision
from shapely import wkt

from em1.likelihood import log_lhood_comp, log_lhood_comp_FF, posterior, prior
from em1.utils import daterange, detect_shut_down


class EM1():

    def __init__(self, model_input: ModelInput, **kwargs):
        # set up logger
        set_up_logging(__name__)
        self._logger = logging.getLogger(__name__)

        self._logger.info('Model Initialized')

        self.model_input = model_input
        self.parameters = \
            model_input.model_parameters | model_input.model_settings

        # last time injection happened (not necessarily shut-down)
        self.t_shut = 0.
        # total injected volume until detected shut-down (if any)
        self.Q_shut = 0.
        self.N_shut = 0.  # number of events during injection
        self.tau = self.parameters['Tau']
        self.PostRate = 0.
        self.past_shut = False
        self.tau_force = self.parameters['tau_force']
        self.mode = self.parameters['mode']
        self.column_flow = 'topflow'

        self.af = self.parameters['afb']
        self.b_value = self.parameters['b_value']
        self.Mc = self.parameters['Mc']
        self.dM = self.parameters['dM']
        self.Mc_corr = self.parameters.get('mc_corr', 0)
        self.Sigma = 0
        self.postdecay = 0
        self.Nsim = self.parameters['Nsim']
        section_id = self.parameters['well_section_id']

        self.bounding_polygon = wkt.loads(model_input.bounding_polygon)
        self.depth_min = model_input.depth_min
        self.depth_max = model_input.depth_max

        self.forecast_start = model_input.forecast_start
        self.forecast_end = model_input.forecast_end

        self.hydraulics = BoreholeHydraulics(
            model_input.injection_observation[0])

        self.scenario = BoreholeHydraulics(model_input.injection_plan[0])

        self.seismicity = Catalog.from_quakeml(
            model_input.seismicity_observation, include_quality=True)

        # get dataframes from HYDSJONParser
        self.hydraulic_df = self.hydraulics[section_id].hydraulics
        self.hydraulic_df = self.hydraulic_df[
            ~self.hydraulic_df['topflow'].isna()]
        self.hydraulic_df = self.hydraulic_df[
            ~self.hydraulic_df['toppressure'].isna()]

        self.scenario_df = self.scenario[section_id].hydraulics

        self.observation_start = self.hydraulic_df.index.min()
        self.observation_end = np.array(
            self.forecast_start).astype('datetime64')

        self.hydraulic_df = \
            self.hydraulic_df[self.hydraulic_df.index < self.observation_end]
        if 'topflow' in self.hydraulic_df.columns:
            self.hydraulic_df['topflow'] = \
                self.hydraulic_df['topflow'].replace(0, 1e-7)

        self.seismicity = self.seismicity[self.seismicity['time']
                                          < self.forecast_start]
        self.seismicity = self.seismicity[self.seismicity['time']
                                          >= self.observation_start]

        self.seismicity = self.seismicity[~self.seismicity['magnitude'].isna()]
        self.seismicity['magnitude'] = bin_to_precision(
            self.seismicity['magnitude'], self.dM)


        if 'associatedphasecount' in self.seismicity.columns:
            self.seismicity['associatedphasecount'] = \
                self.seismicity['associatedphasecount'].astype(float)
            # is this sensible?
            self.seismicity['associatedphasecount'] = \
                self.seismicity['associatedphasecount'].fillna(0)

            self.seismicity = \
                self.seismicity[self.seismicity['associatedphasecount']
                                >= self.parameters['n_phases']]


        # add ts column to seismicity dataframe
        self.seismicity['ts'] = self.seismicity.time.map(
            lambda x: (x - self.forecast_start).total_seconds())

        # get the forecast times
        self.forecast_moments = daterange(self.forecast_start,
                                          self.forecast_end,
                                          self.parameters['epoch_duration'],
                                          self.observation_start)

        self.scenario_df = self.scenario_df[self.scenario_df.index
                                            <= self.forecast_end]
        self.scenario_df = self.scenario_df[self.scenario_df.index
                                            > self.forecast_start]

        sc_all_negative = (self.scenario_df['topflow'] < 0).all()
        sc_all_identical = self.scenario_df['topflow'].nunique() == 1

        if sc_all_negative and sc_all_identical:
            self.scenario_df["topflow"] = \
                (-self.scenario_df["topflow"].values[0]
                 * np.ones(len(self.scenario_df))
                 * self.hydraulic_df["topflow"].values[-1])

    def run(self):
        """
        Run the model. Calls self.fit() and returns the result of
        self.predict().
        """
        self.fit()
        results = self.predict()

        results = self.parse_results(results)

        return results

    def fit(self):
        self._logger.info('Model Fitting')

        # fit to the training data
        self.dt_fit = np.array(
            (self.hydraulic_df.index.values
             - self.hydraulic_df.index.values[0]).tolist()) / 1.e9

        dV = np.maximum(self.hydraulic_df[self.column_flow].values, 0)
        dV[dV < 0] = 0

        self.t_shut, self.Q_shut, self.PostRate = detect_shut_down(
            dV, self.dt_fit)

        mag = self.seismicity['magnitude']
        # number of simulation for the paramters within CI
        Nsim = self.Nsim
        # bound for CI first MLE then Bayes
        CIlow = .025
        CIhigh = .975
        # Critical value of CHI^2(1dof) at 95% = 100(1-0.05)% --> test at
        # alpha=0.05
        Chi2crt = 3.84
        # Chi2crt = 10.828  # Critical value of CHI^2(1dof) at 99.9% =
        # 100(1-0.05)% --> test at alpha=0.05

        # Mc estimation if not given as input
        if self.Mc is None:
            #Mrange = np.around(np.arange(np.amax(mag) + 5 * self.dM, np.amin(
            #    mag) - 5 * self.dM, -self.dM), decimals=int(-np.log(self.dM)))
            #argMc = np.argmax(stats.gaussian_kde(mag)(Mrange))
            #self.Mc = Mrange[argMc]
            Mc = estimate_mc_maxc(self.seismicity['magnitude'],
                                  delta_m=self.dM,
                                  correction_factor=self.Mc_corr)
            #Mc, b_value = Calibrate_Observed_GR(
            #    self.seismicity, self.dM, 'magnitude')
            self.Mc = Mc



        # selecting catalog above completness
        IS_Catalog = self.seismicity[self.seismicity['magnitude'] >= self.Mc]
        self.N_shut = len(IS_Catalog[IS_Catalog['ts'] <= self.t_shut])

        listOfStrings = ['freq', 'MLE', 'BayesMAP']
        if self.mode not in listOfStrings:
            self._logger.warning(
                f'Mode "{self.mode}" not found. '
                f'Possible values are {listOfStrings} '
                f'Using default mode (MLE).')
            self.mode = 'MLE'

        # calculate prior independently of the mode
        self.prior = prior(norm=True)

        if self.mode == 'freq':
            # Maximum Likelihood estimation b-value
            #self.b_value = np.log10(
            #    np.exp(1.)) / (np.mean(mag[mag >= (self.Mc)])
            #                   - (self.Mc - self.dM / 2))
            self.b_value = round(
                estimate_b(self.seismicity['magnitude'].values[self.seismicity['magnitude'].values >= self.Mc],
                           self.Mc, delta_m=self.dM), 2)
            # af estimation
            self.af = self.b_value * self.Mc + \
                np.log10(self.N_shut) - np.log10(self.Q_shut)
            # Seismogenic index
            # self.Sigma = 10 ** (self.af - self.b_value * self.Mc)

        elif self.mode == 'MLE':
            Mw = IS_Catalog.magnitude.values
            Ts = IS_Catalog.ts.values
            N = len(IS_Catalog['ts'])

            initial_theta = (self.af, self.b_value, self.tau)
            args = (N, self.Mc, dV, self.dt_fit, Mw, Ts,
                    self.t_shut, self.N_shut, self.Q_shut, self.PostRate)

            result = opt.fmin(
                func=log_lhood_comp,
                x0=initial_theta,
                args=args,
                maxfun=1e4,
                full_output=1)

            self.af = result[0][0]
            self.b_value = result[0][1]
            self.tau_cal = result[0][2]
            LLopt = result[1]

            # confidence interval for af
            nbin = 1000  # 3 * len(self.prior[3])  # reduces the bin size by 3
            if self.af < np.min(self.prior[3]):
                afgrid = np.linspace(self.af - 3, self.af + 3, num=nbin)
                self._logger.warning('af smaller than prior range')
            elif self.af > np.max(self.prior[3]):
                afgrid = np.linspace(self.af - 3, self.af + 3, num=nbin)
                self._logger.warning('af larger than prior range')
            else:
                afgrid = np.linspace(
                    np.min(
                        self.prior[3]), np.max(
                        self.prior[3]), num=nbin)

            prrng = []
            LLrng = []
            for pr in afgrid:

                grid_theta = (pr, self.b_value, self.tau_cal)
                LLval = log_lhood_comp(
                    grid_theta,
                    N,
                    self.Mc,
                    dV,
                    self.dt_fit,
                    Mw,
                    Ts,
                    self.t_shut,
                    self.N_shut,
                    self.Q_shut,
                    self.PostRate)
                LLrng.append(LLval)
                # likelihood ratio test using Chi square distrubution of
                # 2(l(theta_hat) - l(theta))<=Chi^2(1-alpha)
                if np.abs(LLopt - LLval) < Chi2crt / 2:
                    prrng.append(pr)

            # self.LLrngaf = np.array(LLrng)
            prrng = np.array(prrng)
            afcimin = np.min(prrng)
            afcimax = np.max(prrng)
            self.afCI = [afcimin, afcimax]
            sigmaaf = (afcimax - afcimin) / 4
            self.af_dstr = stats.norm.rvs(
                loc=self.af, scale=sigmaaf, size=Nsim)

            # confidence interval for b_value
            bvalgrid = np.linspace(
                np.min(
                    self.prior[4]), np.max(
                    self.prior[4]), num=nbin)
            if self.b_value < np.min(self.prior[4]):
                bvalgrid = np.linspace(
                    self.b_value - 0.5, self.b_value + 0.5, num=nbin)
                self._logger.warning('b_value smaller than prior range')
            elif self.b_value > np.max(self.prior[4]):
                bvalgrid = np.linspace(
                    self.b_value - 0.5, self.b_value + 0.5, num=nbin)
                self._logger.warning('b_value larger than prior range')
            else:
                bvalgrid = np.linspace(
                    np.min(
                        self.prior[4]), np.max(
                        self.prior[4]), num=nbin)

            prrng = []
            LLrng = []
            for pr in bvalgrid:

                grid_theta = (self.af, pr, self.tau_cal)
                LLval = log_lhood_comp(
                    grid_theta,
                    N,
                    self.Mc,
                    dV,
                    self.dt_fit,
                    Mw,
                    Ts,
                    self.t_shut,
                    self.N_shut,
                    self.Q_shut,
                    self.PostRate)
                LLrng.append(LLval)
                if np.abs(LLopt - LLval) < Chi2crt / 2:
                    prrng.append(pr)

            # self.LLrngbval = np.array(LLrng)
            prrng = np.array(prrng)
            b_valcimin = np.min(prrng)
            b_valcimax = np.max(prrng)
            self.b_valueCI = [b_valcimin, b_valcimax]
            sigmab = (b_valcimax - b_valcimin) / 4
            self.b_value_dstr = stats.norm.rvs(
                loc=self.b_value, scale=sigmab, size=Nsim)

            # confidence interval for tau # the LL of tau is flat so it return
            # all the value in prior[tau]
            taugrid = np.linspace(
                np.min(
                    self.prior[5]), np.max(
                    self.prior[5]), num=nbin)

            prrng = []
            LLrng = []
            for pr in taugrid:

                grid_theta = (self.af, self.b_value, pr)
                LLval = log_lhood_comp(
                    grid_theta,
                    N,
                    self.Mc,
                    dV,
                    self.dt_fit,
                    Mw,
                    Ts,
                    self.t_shut,
                    self.N_shut,
                    self.Q_shut,
                    self.PostRate)
                LLrng.append(LLval)
                if np.abs(LLopt - LLval) < Chi2crt / 2:
                    prrng.append(pr)
            # self.LLrngTau= np.array(LLrng)
            prrng = np.array(prrng)
            taucimin = np.min(prrng)
            taucimax = np.max(prrng)
            self.tauCI = [taucimin, taucimax]
            sigmatau = (taucimax - taucimin) / 4
            self.tau_dstr = stats.norm.rvs(
                loc=self.tau_cal, scale=sigmatau, size=Nsim)

        elif self.mode == 'BayesMAP':
            # Calcualte prior based on previous experiments
            # self.prior = prior(norm=True) # move up out of the if

            Mw = IS_Catalog.magnitude.values
            Ts = IS_Catalog.ts.values
            N = len(IS_Catalog['ts'])

            A0 = np.tile(self.prior[3], (len(self.prior[5]), 1)).T
            T0 = np.tile(self.prior[5], (len(self.prior[3]), 1))
            II = np.ones((len(self.prior[5]), len(self.prior[3]))).T

            theta_range = (A0, T0, self.prior[4], II)

            args = (N, self.Mc, dV, self.dt_fit, Mw, Ts,
                    self.t_shut, self.N_shut, self.Q_shut, self.PostRate)

            # Calculate log likelihood distribution
            LL = log_lhood_comp_FF(theta_range, args)

            # calculate a posteriori distribution

            self.posterior = posterior(self.prior, LL)

            # parameters estimated as maximum a posteriori
            self.b_value = self.prior[4][np.where(
                self.posterior == np.max(self.posterior))[2]][0]
            # af estimation
            self.af = self.prior[3][np.where(
                self.posterior == np.max(self.posterior))[0]][0]
            # tau estimate
            self.tau_cal = self.prior[5][np.where(
                self.posterior == np.max(self.posterior))[1]][0]

            # CI for parameters estimated as maximum a posteriori (there might
            # be a cleaver and shorter way to do that)

            # b_value CI estimation
            Biv_pdf = np.sum(self.posterior, axis=1)
            Uni_pdf = np.sum(Biv_pdf, axis=0)
            # self.b_pdf = Uni_pdf
            Uni_CDF = np.cumsum(Uni_pdf / np.sum(Uni_pdf))
            idx_low = np.argwhere(Uni_CDF < CIlow)[-1]
            idx_high = np.argwhere(Uni_CDF > CIhigh)[0]
            var_low = self.prior[4][idx_low + 1]
            var_high = self.prior[4][idx_high - 1]
            self.b_valueCI = [var_low, var_high]
            sigmab = (var_high - var_low) / 4
            self.b_value_dstr = stats.norm.rvs(
                loc=self.b_value, scale=sigmab, size=Nsim)

            # af CI estimation
            Biv_pdf = np.sum(self.posterior, axis=1)
            Uni_pdf = np.sum(Biv_pdf, axis=1)
            # self.af_pdf = Uni_pdf
            Uni_CDF = np.cumsum(Uni_pdf / np.sum(Uni_pdf))
            idx_low = np.argwhere(Uni_CDF < CIlow)[-1]
            idx_high = np.argwhere(Uni_CDF > CIhigh)[0]
            var_low = self.prior[3][idx_low + 1]
            var_high = self.prior[3][idx_high - 1]
            self.afCI = [var_low, var_high]
            sigmaaf = (var_high - var_low) / 4
            self.af_dstr = stats.norm.rvs(
                loc=self.af, scale=sigmaaf, size=Nsim)

            # tau CI estimation -->> warning spread pdf
            Biv_pdf = np.sum(self.posterior, axis=0)
            Uni_pdf = np.sum(Biv_pdf, axis=1)
            Uni_CDF = np.cumsum(Uni_pdf / np.sum(Uni_pdf))
            self.Taucdf = Uni_CDF
            idx_low = np.argwhere(Uni_CDF < CIlow)[-1]
            idx_high = np.argwhere(Uni_CDF > CIhigh)[0]
            var_low = self.prior[5][idx_low + 1]
            var_high = self.prior[5][idx_high - 1]
            self.tauCI = [var_low, var_high]
            sigmatau = (var_high - var_low) / 4
            self.tau_dstr = stats.norm.rvs(
                loc=self.tau_cal, scale=sigmatau, size=Nsim)

        # Seismogenic index
        self.Sigma = 10 ** (self.af - self.b_value * self.Mc)
        self.Sigma_dstr = 10 ** (self.af_dstr - self.b_value_dstr * self.Mc)
        self._logger.info('Estimated parameters: '
                          + str(self.af)
                          + ', '
                          + str(self.b_value)
                          + ', '
                          + str(self.Sigma))

        B = len(IS_Catalog) - self.N_shut
        # self.PostRate = self.Q_shut / self.t_shut
        if (B > 1):
            self.past_shut = True
            # self.PostRate = (
            #     self.N_shut -
            #     len(IS_Catalog[IS_Catalog['ts'] <
            #                    (self.t_shut - 6. * 3600)])) / 6. * 3600

        if (B >= 3):  # accept that shut-down exists in Well Logs
            if not self.tau_force:
                if self.mode == 'freq':
                    t_end = (IS_Catalog['ts'].max() - self.t_shut)
                    r = -t_end / B
                    self.tau = 1000 * (t_end * B) / \
                        (B * lambertw(r * np.exp(r)).real + t_end)
                elif self.mode == 'MLE':
                    if self.tau_cal > 0:
                        self.tau = self.tau_cal
                elif self.mode == 'BayesMAP':
                    if self.tau_cal > 0:
                        self.tau = self.tau_cal

        # Recalculate the model cumulative seismicity to compare with the data
        Q_measured = np.cumsum(
            dV
            * np.diff(
                np.append(
                    self.dt_fit[0],
                    self.dt_fit)))
        y = Q_measured * self.Sigma
        MetaSeismic = np.where(self.dt_fit > self.t_shut)
        if len(MetaSeismic[0]) > 0:
            past = y[MetaSeismic[0][0] - 1]
        else:
            past = 0
        meta_t = self.dt_fit - self.t_shut
        y[MetaSeismic] = (-np.exp(-meta_t[MetaSeismic] / self.tau)
                          + 1.) * self.PostRate * self.tau * self.Sigma + past
        # in the case where you have 2 decay phases...e.g. calibrating one
        # injection and forecasting a second one
        if self.past_shut:
            self.postdecay = y[-1] - past

        self.y_fit = y

    def predict(self):
        self._logger.info('Model Predicting')

        # dt = self.scenario_df.time.values
        self.dt_for = np.array(
            (self.scenario_df.index.values
             - self.hydraulic_df.index.values[0]).tolist()) / 1.e9
        dV = np.maximum(self.scenario_df[self.column_flow].values, 0)
        # the fitted t_shut and Q_shut might not be the ending of the
        # stimulation so find the one implied by the planned strategy
        t_shut, Q_shut, PostRate = detect_shut_down(
            np.append(self.Q_shut / self.t_shut, dV),
            np.append(self.t_shut, self.dt_for))

        # N_shut = self.Sigma * Q_shut

        # Sigma = 10**(self.af - self.b_value * self.Mc)

        Q_forecasted = np.interp(
            self.forecast_moments,
            self.dt_for,
            np.cumsum(dV * np.diff(np.append(
                self.dt_for[0],
                self.dt_for))))

        # if (self.past_shut):  # tau has been calibrated, shutting was found
        #     t_shut, Q_shut, PostRate = [
        #         self.t_shut, self.Q_shut, self.PostRate]
        #     N_shut = 0

        # y = (Q_forecasted + self.Q_shut) * self.Sigma +self.postdecay
        y = (Q_forecasted + self.Q_shut) * self.Sigma
        y_grd = np.zeros((len(y), self.Nsim))
        for i in range(self.Nsim):
            y_grd[:, i] = (Q_forecasted + self.Q_shut) * self.Sigma_dstr[i]

        self._logger.info(f"EM1 considers that injection stops at {t_shut}s "
                          f"and a total of {Q_forecasted[-1]} m3 will have "
                          "to be injected.")

        # Continue calculating decay curve if past shut-in
        if self.past_shut:
            MetaSeismic = np.where(self.forecast_moments >= self.t_shut)
            if len(MetaSeismic[0]) > 0:
                past = y[MetaSeismic[0][0]]
                past_grd = y_grd[MetaSeismic[0][0], :]
            else:
                past = 0
                past_grd = y_grd[0, :] * 0

            meta_t = self.forecast_moments - self.t_shut
            y[MetaSeismic] = (
                -np.exp(-meta_t[MetaSeismic] / self.tau) + 1.) * \
                self.PostRate * self.tau * self.Sigma + \
                past + y[MetaSeismic] - y[0]

            for i in range(self.Nsim):
                y_grd[MetaSeismic, i] = (
                    -np.exp(-meta_t[MetaSeismic] / self.tau) + 1.) * \
                    self.PostRate * self.tau * self.Sigma_dstr[i] + \
                    past_grd[i] + y_grd[MetaSeismic, i] - y_grd[0, i]

            # Adding now additional decay
            MetaSeismic2 = np.where(self.forecast_moments >= t_shut)
            if len(MetaSeismic2[0]) > 0:
                past2 = y[MetaSeismic2[0][0]]
                past2_grd = y_grd[MetaSeismic2[0][0], :]
            else:
                past2 = 0
                past2_grd = y_grd[0, :] * 0

            meta_t2 = self.forecast_moments - t_shut
            y[MetaSeismic2] = (-np.exp(-meta_t2[MetaSeismic2] / self.tau)
                               + 1.) * PostRate * self.tau * self.Sigma + past2

            for i in range(self.Nsim):
                y_grd[MetaSeismic2, i] = (
                    -np.exp(-meta_t2[MetaSeismic2] / self.tau) + 1.) * \
                    self.PostRate * self.tau * self.Sigma_dstr[i] + \
                    past2_grd[i]

            # Sigma = 10**(self.af - self.b_value * self.Mc)

            # meta_t = Forecasts_Moments - t_shut
            # MetaSeismic = np.where(Forecasts_Moments > t_shut)

            # y =(Sigma * Q_forecasted/Forecasts_Moments)
            # y[MetaSeismic] = (-np.exp(-meta_t[MetaSeismic] /
            #                   self.tau) + 1.) * self.PostRate * self.tau

        NN = ((np.max(self.dt_fit) - np.min(self.dt_fit))
              / self.parameters['epoch_duration']).astype(int) + 1
        self._time_fit = np.linspace(
            np.min(self.dt_fit), np.max(self.dt_fit), NN)
        # rate_fit = np.zeros((len(self._time_fit[1:]),self.Nsim))
        rate_fit = np.diff(np.interp(self._time_fit, self.dt_fit, self.y_fit))

        N_observed = len(
            self.seismicity[self.seismicity['magnitude'] >= self.Mc])

        if self.Nsim > 1:
            EM1_alpha = self.af_dstr
            EM1_b = self.b_value_dstr
            EM1_forecast = np.diff(y_grd, axis=0)
            EM1_a = np.log10(np.cumsum(EM1_forecast, axis=0)
                             + N_observed + 1e-20) + EM1_b * self.Mc
        else:
            EM1_alpha = self.af
            EM1_b = self.b_value
            EM1_forecast = np.diff(y).reshape(
                (len(self.forecast_moments) - 1, 1))
            EM1_a = (np.log10(np.cumsum(EM1_forecast, acis=0)
                              + N_observed + 1e-20) + EM1_b
                     * self.Mc).reshape((len(self.forecast_moments) - 1, 1))

        # result:
        #        Nfit is [max(time_data)-min(time_data)]/epoch_duration +1
        #        Nt is len(time_forecast)-1 (i.e. diff(time_forecast))
        #
        # 0       b_value distribution ->
        #               1*Nsim (up to Nx * Ny * Nz * Nt * Nsim potentially)
        # 1       Mc  -> 1 (up to Nx * Ny * Nz * Nt * Nsim potentially)
        # 2       a_value distribution ->
        #               Nt*Nsim (up to Nx * Ny * Nz * Nt * Nsim potentially)
        # 3       alpha distribution -> Nt*Nsim
        # 4       rate_fit -> 1*Nfit (up to Nx * Ny * Nz * Nfit potentially)
        # 5       rate_forecast -> Nt*Nsim (up to Nx * Ny * Nz * Nt * Nsim
        # potentially)

        return EM1_a, EM1_alpha, EM1_b, rate_fit, EM1_forecast

    def parse_results(self, results):
        """
        Parse Results to a ForecastGRRateGrid per timestep.
        A ForecastGRRateGrid requires at least the following columns:
            'number_events', 'longitude_min', 'longitude_max',
            'latitude_min', 'latitude_max', 'depth_min', 'depth_max',
            'a', 'b', 'mc', 'grid_id'

        """
        fit_datetimes = [self.observation_start + timedelta(seconds=t)
                         for t in self._time_fit]
        forecast_datetimes = [self.observation_start + timedelta(seconds=t)
                              for t in self.forecast_moments]

        bounds = list(self.bounding_polygon.bounds)
        bounds.insert(5, self.depth_min)
        bounds.insert(2, self.depth_max)

        parsed_results = []
        for i in range(len(fit_datetimes) - 1):
            rate = {'number_events': [results[3][i]],
                    'longitude_min': bounds[0],
                    'longitude_max': bounds[3],
                    'latitude_min': bounds[1],
                    'latitude_max': bounds[4],
                    'depth_min': bounds[5],
                    'depth_max': bounds[2],
                    # self.af,
                    'a': np.log10(len(
                        self.seismicity[self.seismicity['magnitude']
                                        > self.Mc])) + self.b_value * self.Mc,
                    'alpha': self.af,
                    'b': self.b_value,
                    'mc': self.Mc,
                    'grid_id': 0}

            # Create a ForecastGRRateGrid for each Timestep
            rategrid = ForecastGRRateGrid(rate,
                                          starttime=fit_datetimes[i],
                                          endtime=fit_datetimes[i + 1])

            parsed_results.append(rategrid)

        for i in range(len(forecast_datetimes) - 1):
            rate = {'number_events': results[4][i],
                    'longitude_min': [bounds[0]] * self.Nsim,
                    'longitude_max': [bounds[3]] * self.Nsim,
                    'latitude_min': [bounds[1]] * self.Nsim,
                    'latitude_max': [bounds[4]] * self.Nsim,
                    'depth_min': [bounds[5]] * self.Nsim,
                    'depth_max': [bounds[2]] * self.Nsim,
                    'a': results[0][i],
                    'alpha': results[1],
                    'b': results[2],
                    'mc': [self.Mc] * self.Nsim,
                    'grid_id': list(range(self.Nsim))}

            # Create a ForecastGRRateGrid for each Timestep
            rategrid = ForecastGRRateGrid(rate,
                                          starttime=forecast_datetimes[i],
                                          endtime=forecast_datetimes[i + 1])

            parsed_results.append(rategrid)

        return parsed_results
