import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

def detect_shut_down(dV, dt):
    V = np.cumsum(np.maximum(dV[1:], 0) * np.diff(dt))
    t_shut = dt[np.argmax(V) + 1]
    Q_shut = np.amax(V)
    PostRate = dV[np.argmax(V) + 1]
    return t_shut, Q_shut, PostRate


def daterange(start_date,
              end_date,
              epoch_duration=None,
              training_start_date=None):
    """Numpy array returned with forecasting intervals in seconds
    increasing with constant increment, inclusive of the end_date.

    :param st art_date: Datetime for generator to start with.
    :param end_date: Datetime for generator to finish with.
    :param epoch_duration: Number of seconds between returned values.

    :rtype: np.array of seconds from training start time to forecast for
    """
    total_seconds = (end_date - start_date).total_seconds()
    if not epoch_duration:
        epoch_duration = total_seconds

    # the +1 is required to match the r code which gives the range
    # including the final variable.
    daterange_list = [
        i * epoch_duration for i
        in range(int(total_seconds / epoch_duration) + 1)]
    if training_start_date:
        daterange_list = [i + (
            start_date - training_start_date).total_seconds()
            for i in daterange_list]
    date_range_array = np.array(daterange_list)
    return date_range_array


def plotting_data(model):
    fig, ax = plt.subplots(1, 2, figsize=(20, 6))

    ax[0].scatter(model.seismicity.time, (model.seismicity['magnitude'] - np.min(model.seismicity['magnitude']))
                  / (np.max(model.seismicity['magnitude'] - np.min(model.seismicity['magnitude']))) / 4,
                  color='lightgrey', alpha=.5, zorder=2, label='seismics')
    ax[0].step(model.hydraulic_df.index, model.hydraulic_df['topflow']
               / np.max(model.hydraulic_df['topflow']) / 4, color='darkblue', label='hydraulics[topflow]')
    ax[0].step(model.hydraulic_df.index, model.hydraulic_df['toppressure']
               / np.max(model.hydraulic_df['toppressure']) / 4, color='darkred', label='hydraulics[toppressure]')
    ax[0].set_xlabel('Date')
    ax[0].set_ylabel('Normalised inputs')
    ax[0].set_title('Training data')
    ax[0].legend(loc='upper left')
    ax[0].spines['right'].set_visible(False)
    ax[0].spines['top'].set_visible(False)

    ax[1].plot(model.hydraulic_df.index, model.hydraulic_df['topflow'], color='darkblue',
               label='hydraulics[topflow]')
    ax[1].plot(model.scenario_df.index, model.scenario_df[model.column_flow],
               color='royalblue', label='scenario')
    ax[1].set_xlabel('Date')
    ax[1].set_ylabel('Flow rate [m$^3$/s]')
    ax[1].set_title('Flow time series')
    ax[1].legend()
    ax[1].spines['right'].set_visible(False)
    ax[1].spines['top'].set_visible(False)
    fig.tight_layout()


def plotting_results(model, model_output, seism_df=pd.DataFrame()):

    inj_t = np.array((model.hydraulic_df.index.values
                     - model.hydraulic_df.index.values[0]).tolist()) / 1.e9
    Dtau = model.model_input.model_settings['epoch_duration']

    NN = ((inj_t[-1] - inj_t[0])
          / Dtau).astype(int) + 1
    time_fit = np.linspace(
        inj_t[0], inj_t[-1], NN)

    time_fit_mids = time_fit[1:]

    time_forecast_mids = model.forecast_moments[1:] - Dtau / 2

    seism = model.seismicity
    seism = seism[seism['magnitude'] >= model.Mc]
    seism.sort_values(by='time', inplace=True)
    seism_t = np.array(
        (seism['time'].values - model.hydraulic_df.index.values[0]).tolist()) / 1.e9

    # range pred

    y_pred_mean = np.median(model_output[4], axis=1)
    up = np.percentile(model_output[4], axis=1, q=75)
    down = np.percentile(model_output[4], axis=1, q=25)
    down[down < 0] = 0

    if not seism_df.empty:
        seism_t_all = np.array(
            (seism_df['time'].values - model.hydraulic_df.index.values[0]).tolist()) / 1.e9
        ind = np.logical_and(seism_t_all > 0, seism_t_all
                             < model.forecast_moments[-1])
        seism_df_filt = seism_df[ind]
        seism_t_all = seism_t_all[ind]
        seism_t_all_aboveMc = seism_t_all[seism_df_filt['magnitude'] >= model.Mc]
        taubins_all = np.arange(0, model.forecast_moments[-1] + Dtau, Dtau)
        tmids_all = taubins_all[1:] - Dtau / 2
        rate_all = np.histogram(seism_t_all_aboveMc, bins=taubins_all)

    ##
    fig, ax = plt.subplots(1, 2, figsize=(20, 6))

    ax[0].scatter(seism_t, np.arange(1, len(seism_t) + 1),
                  color='darkgrey', label='observations')
    ax[0].plot(time_fit_mids, np.cumsum(model_output[3]),
               color='darkgreen', label='model fit')
    ax[0].set_xlabel('Time increment [s]')
    ax[0].set_ylabel('Cumulative number of events')
    ax[0].set_title('Training fit')
    ax[0].legend()
    ax[0].spines['right'].set_visible(False)
    ax[0].spines['top'].set_visible(False)

    # ax[1].plot(model.res_fit_t, model.rate_train, color = 'lightgrey', label = 'observations')
    ax[1].plot(time_fit_mids, model_output[3],
               color='darkgreen', label='model fit')
    ax[1].plot(time_forecast_mids, y_pred_mean,
               color='darkred', label='model prediction')
    ax[1].scatter(time_forecast_mids, y_pred_mean, color='darkred')
    ax[1].fill_between(time_forecast_mids, down, up,
                       facecolor='darkred', alpha=.3)
    if not seism_df.empty:
        ax[1].step(tmids_all, rate_all[0], where='mid',
                   label='data', color='darkgrey')
    # ax[1].fill_between(tmids_all, 0, rate_all[0], color='grey', alpha = .5)
    # ax[1].set_xlim(0, 50000)
    # ax[1].set_ylim(0, 150)
    ax[1].set_xlabel('Time increment [s]')
    ax[1].set_ylabel('Event rate [time bin]')
    ax[1].set_title('Event-rate time series')
    ax[1].legend(loc='upper left')
    ax[1].spines['right'].set_visible(False)
    ax[1].spines['top'].set_visible(False)
    # fig.tight_layout()
