import logging
import os
from datetime import datetime

import numpy as np
import pandas as pd
from hydws.parser import BoreholeHydraulics
from seismostats import Catalog
from seismostats.utils.coordinates import CoordinateTransformer
from shapely import Polygon, to_wkt

from em1.hermes import ModelInput
from em1.model import EM1
from em1.utils import plotting_data, plotting_results

# Set up logger
date_strftime_format = "%d-%m-%y %H:%M:%S"
logging.basicConfig(
    format='%(asctime)s %(levelname)s  %(message)s',
    datefmt=date_strftime_format,
    level=logging.INFO)


def main():
    ############################################################
    # HYDRAULICS
    dirpath = os.path.dirname(os.path.abspath(__file__))
    Data_Path = os.path.join(dirpath, "data/")
    hydraulics = pd.read_hdf(Data_Path + 'TrainingData.h5')

    hydraulics_plan = pd.read_hdf(Data_Path + 'Scenario.h5')

    # define start and end of forecast
    forecast_start = datetime.utcfromtimestamp(
        hydraulics_plan.index.values[0].tolist() / 1e9)
    forecast_end = datetime.utcfromtimestamp(
        hydraulics_plan.index.values[-1].tolist() / 1e9)

    # rename columns to match hydws
    hydraulics = hydraulics.drop('time', axis=1).rename(
        columns={'pressure': 'toppressure'})
    hydraulics_plan.index.names = ['datetime']
    hydraulics_plan["topflow"] = -1 * np.ones((len(hydraulics_plan), 1))

    # generate empty parser and load dataframe to empty section
    parser = BoreholeHydraulics()
    section_id = parser.section_from_dataframe(hydraulics)
    hydjson = parser.to_json()

    # replace hydraulics with scenario hydraulics
    parser[section_id].hydraulics = hydraulics_plan
    scenario_hydjson = parser.to_json()

    ############################################################
    # SEISMICITY

    proj_string = 'epsg:2056'

    # define coordinate transformer, local <-> swiss, external <-> WGS84
    transformer = CoordinateTransformer(proj_string)
    top_section_swiss = transformer.to_local_coords(
        8.47449792444771,
        46.5098187019071,
        1271.43402303251
    )

    # CoordinateTransformer with reference point in epsg:2056
    transformer = CoordinateTransformer(proj_string, *top_section_swiss)

    # Read seismics, transform to WGS84
    catalog_df = pd.read_hdf(Data_Path + 'TrainingTarget.h5')
    catalog_df['Xs'], catalog_df['Ys'], catalog_df['Zs'] = \
        transformer.from_local_coords(catalog_df.Xs.values,
                                      catalog_df.Ys.values,
                                      catalog_df.Zs.values)

    catalog_df = catalog_df.reset_index().rename(
        columns={
            'Xs': 'longitude',
            'Ys': 'latitude',
            'Zs': 'depth',
            'mag': 'magnitude',
            'datetime': 'time'})

    catalog = Catalog(catalog_df)
    catalog['magnitude_type'] = 'mag'

    quakeml = catalog.to_quakeml()

    bounding_polygon = Polygon([[-150, -150], [-150, 150], [150, 150],
                               [150, -150], [-150, -150]])
    bounding_polygon = transformer.polygon_from_local_coords(bounding_polygon)

    data = {
        'forecast_start': forecast_start,
        'forecast_end': forecast_end,
        'injection_observation': [hydjson],
        'injection_plan': [scenario_hydjson],
        'depth_max': -150 + top_section_swiss[2],
        'depth_min': 150 + top_section_swiss[2],
        'bounding_polygon': to_wkt(bounding_polygon),
        'seismicity_observation': quakeml,
        'model_parameters': {
            'b_value': 1,
            'afb': -2,
            'Tau': 60,
            'dM': 0.05,
            'Mc': None,
            'tau_force': False,
            'Nsim': 100,
            'mode': 'MLE'
        },
        'model_settings': {
            'well_section_id': str(section_id),
            'n_phases': 8,
            'epoch_duration': 600
        }
    }

    model_input = ModelInput(**data)

    model = EM1(model_input)

    plotting_data(model)

    model.fit()

    results = model.predict()

    # results = model.parse_results(results)
    catalog_df_full = pd.read_hdf(Data_Path + 'seismic_data_full_located.h5')
    catalog_df_full.rename(columns={"X": "Xs", "Y": "Ys", "Depth": "Zs",
                           "MomMag": "magnitude", "Time(UTC)": "time"},
                           inplace=True)
    catalog_df_full = catalog_df_full[catalog_df_full["magnitude"] > -5]
    catalog_df_full.set_index("time", inplace=True)
    catalog_df_full = catalog_df_full[catalog_df_full.index
                                      >= hydraulics.index.values[0]]
    catalog_df_full = catalog_df_full[catalog_df_full.index <= forecast_end]
    catalog_df_full = catalog_df_full.reset_index()

    plotting_results(model, results, catalog_df_full)

    return model


if __name__ == "__main__":
    result = main()
