import logging
import os
from datetime import datetime

import pandas as pd
from hydws.parser import BoreholeHydraulics
from seismostats import Catalog
from seismostats.utils.coordinates import CoordinateTransformer
from shapely import Polygon, to_wkt

from em1.hermes import run_em1, run_em1_validated  # noqa

# Set up logger
date_strftime_format = "%d-%m-%y %H:%M:%S"
logging.basicConfig(
    format='%(asctime)s %(levelname)s  %(message)s',
    datefmt=date_strftime_format,
    level=logging.INFO)


def main():
    ############################################################
    # HYDRAULICS
    dirpath = os.path.dirname(os.path.abspath(__file__))
    Data_Path = os.path.join(dirpath, "data/")
    hydraulics = pd.read_hdf(Data_Path + 'TrainingData.h5')

    hydraulics_plan = pd.read_hdf(Data_Path + 'Scenario.h5')

    # define start and end of forecast
    forecast_start = datetime.utcfromtimestamp(
        hydraulics_plan.index.values[0].tolist() / 1e9)
    forecast_end = datetime.utcfromtimestamp(
        hydraulics_plan.index.values[-1].tolist() / 1e9)

    # rename columns to match hydws
    hydraulics = hydraulics.drop('time', axis=1).rename(
        columns={'pressure': 'toppressure_value',
                 'topflow': 'topflow_value'})
    hydraulics_plan.index.names = ['datetime']

    # generate empty parser and load dataframe to empty section
    parser = BoreholeHydraulics()
    section_id = parser.section_from_dataframe(hydraulics)
    hydjson = parser.to_json()

    # replace hydraulics with scenario hydraulics
    parser[section_id].hydraulics = hydraulics_plan
    scenario_hydjson = parser.to_json()

    ############################################################
    # SEISMICITY

    proj_string = 'epsg:2056'

    # define coordinate transformer, local <-> swiss, external <-> WGS84
    transformer = CoordinateTransformer(proj_string)
    top_section_swiss = transformer.to_local_coords(
        8.47449792444771,
        46.5098187019071,
        1271.43402303251
    )

    # CoordinateTransformer with reference point in epsg:2056
    transformer = CoordinateTransformer(proj_string, *top_section_swiss)

    # Read seismics, transform to WGS84
    catalog_df = pd.read_hdf(Data_Path + 'TrainingTarget.h5')
    catalog_df['Xs'], catalog_df['Ys'], catalog_df['Zs'] = \
        transformer.from_local_coords(catalog_df.Xs.values,
                                      catalog_df.Ys.values,
                                      catalog_df.Zs.values)

    catalog_df = catalog_df.reset_index().rename(
        columns={
            'Xs': 'longitude',
            'Ys': 'latitude',
            'Zs': 'depth',
            'mag': 'magnitude',
            'datetime': 'time'})

    catalog = Catalog(catalog_df)
    catalog['magnitude_type'] = 'Mw'

    quakeml = catalog.to_quakeml()

    bounding_polygon = Polygon([[-150, -150], [-150, 150], [150, 150],
                               [150, -150], [-150, -150]])
    bounding_polygon = transformer.polygon_from_local_coords(bounding_polygon)

    data = {
        'forecast_start': forecast_start,
        'forecast_end': forecast_end,
        'injection_observation': [hydjson],
        'injection_plan': [scenario_hydjson],
        'bounding_polygon': to_wkt(bounding_polygon),  # always in WGS84
        'depth_min': 150 + top_section_swiss[2],
        'depth_max': -150 + top_section_swiss[2],
        'seismicity_observation': quakeml,
        'model_parameters': {
            'b_value': 1,
            'afb': -2,
            'Tau': 60,
            'dM': 0.05,
            'Mc': None,
            'tau_force': False,
            'Nsim': 100,
            'mode': 'MLE'
        },
        'model_settings': {
            'well_section_id': str(section_id),
            'n_phases': 8,
            'epoch_duration': 600
        }
    }

    # # # initialize the model
    # model_output = run_em1(data)  # unvalidated input
    model_output = run_em1_validated(data)  # validated input

    ############################################################
    # results are a list of ForecastGRRateGrids, they can for example
    # be accessed like this:
    print('\n\n'
          f'100 Simulations for timestep from {model_output[-1].starttime} '
          f'until {model_output[-1].endtime}\n',
          model_output[-1])

    # it is easy to add the datetime index to the dataframe:
    print('\n\n'
          'Same dataframe with the datetime index added:\n',
          model_output[-1].add_time_index())

    # as well as concatenating all the simulations into one dataframe:
    full_df = pd.concat([x.add_time_index() for x in model_output])
    print('\n\n'
          'All simulations concatenated into one dataframe:\n',
          full_df)

    ############################################################
    # you have also available here the:

    # input hydraulics (history) as a DataFrame
    print(hydraulics)

    # input injection plan as a DataFrame
    print(hydraulics_plan)

    # input seismicity as a seismostats Catalog
    print(catalog)
    return model_output


if __name__ == "__main__":
    result = main()
